@@ -1,6 +1,7 @@
 # from pandas.core.strings import str_count
 import boto3
 import datetime
+import folium
 import hana_ml
 import io
 import json
@@ -17,6 +18,8 @@ import sys
 import time
 import umap.umap_ as umap
 from fuzzywuzzy import fuzz
+from geopandas import GeoDataFrame
+from geopandas import points_from_xy
 from imblearn.over_sampling import SMOTE
 from imblearn.under_sampling import RandomUnderSampler
 from imblearn.pipeline import Pipeline
@@ -28,6 +31,7 @@ from minio.error import ResponseError
 from numpy import mean, std
 from pytrends.request import TrendReq
 from scipy.cluster.hierarchy import linkage, dendrogram
+from shapely.geometry import mapping, Point, Polygon
 from sklearn.preprocessing import LabelEncoder, MinMaxScaler, PowerTransformer, StandardScaler
 from sklearn.model_selection import train_test_split, cross_val_score, RepeatedStratifiedKFold
 from sklearn.linear_model import LinearRegression, LogisticRegression
@@ -69,6 +73,11 @@ else:
     cred = credf('master')
     credr = credf('redshift')
 
+bucket_name = 'ejgallo-lake-sales-dev'
+benchmark_acv_str = 'prophet/Data_Acquisition/source_data/Archive/2021/benchmark/data/acv_curves.csv'
+benchmark_geo_str = 's3://ejgallo-lake-sales-dev/prophet/Data_Acquisition/source_data/Archive/2021/benchmark/data/geo_codes.csv'
+brand_standards_str = 'prophet/Deployment/Auxillary_Functions/prime_directive/off_brand_standards_Q2.csv'
+upc_tier_str = 'prophet/Data_Acquisition/source_data/Archive/2021/benchmark/data//UPC_Tier_Dict.csv'
 
 ########################### Data Extraction ###########################
 class data_extraction:
@@ -105,7 +114,7 @@ class data_extraction:
         sess = sagemaker.Session()
         region = boto3.session.Session().region_name
         sm = boto3.Session().client('sagemaker')
-        df.to_csv('{}'.format(file_name, index=index))
+        df.to_csv('{}'.format(file_name), index=index)
         data_location = sess.upload_data(path='data', bucket=bucket_name, key_prefix='data')
         print("Uploading is successful...")
 
@@ -122,28 +131,23 @@ class data_extraction:
         s3c.upload_file(Bucket = bucket_name, Filename = 'buffer', Key = KEY)
         print("Uploading is successful...")
 
-#     def lc_reading_in_from_s3(self, bucket_name, file_name, dtypes = None):
-#         s3c = boto3.client(
-#                 's3', 
-#                 region_name = cred.AWS_REGION_NAME,
-#                 aws_access_key_id = cred.AWS_ACCESS_KEY,
-#                 aws_secret_access_key = cred.AWS_SECRET_KEY
-#             )
-#         KEY = '{}'.format(file_name)
-#         obj = s3c.get_object(Bucket=bucket_name, Key = KEY)                         
-#         df = pd.read_csv(io.BytesIO(obj['Body'].read()) , encoding='utf8', low_memory=False, dtype = dtypes) # , on_bad_lines='skip')
-#         return df
-
     def lc_reading_in_from_s3(self, bucket_name, file_name, dtypes = None):
         s3c = self.s3_connection()
         KEY = '{}'.format(file_name)
-        obj = s3c.get_object(Bucket=bucket_name, Key = KEY)
-        streaming_body = obj['Body']
-        content_len = obj['ContentLength']
-        buffered_io = S3StreamingBodyIO(streaming_body._raw_stream, content_len)
-        df = pd.read_csv(io.BytesIO(buffered_io.read()), encoding='utf8', low_memory=False, dtype = dtypes) # on_bad_lines='skip')
+        obj = s3c.get_object(Bucket=bucket_name, Key = KEY)                         
+        df = pd.read_csv(io.BytesIO(obj['Body'].read()) , encoding='utf8', low_memory=False, dtype = dtypes) # , on_bad_lines='skip')
         return df
 
+    # def lc_reading_in_from_s3(self, bucket_name, file_name, dtypes = None):
+    #     s3c = self.s3_connection()
+    #     KEY = '{}'.format(file_name)
+    #     obj = s3c.get_object(Bucket=bucket_name, Key = KEY)
+    #     streaming_body = obj['Body']
+    #     content_len = obj['ContentLength']
+    #     buffered_io = S3StreamingBodyIO(streaming_body._raw_stream, content_len)
+    #     df = pd.read_csv(io.BytesIO(buffered_io.read()), encoding='ISO-8859-1', low_memory=False, dtype = dtypes) # on_bad_lines='skip')
+    #     return df
+
     def lcm_upload_to_s3(self, bucket, file_name):
         minio = self.minio_connection()
         print("Saving... data in {} at {}".format(bucket, file_name))
@@ -202,14 +206,6 @@ class data_extraction:
         results = conn.execute(delete_stmt)
         print("The number of rows deleted are {}...".format(results.rowcount))    
 
-    def lambda_function(self, bucket_name, number):
-        today = datetime.datetime.now()
-        today = today.strftime("%m%d%Y")
-        file_name = "prophet/lambda_cues/lambda_{}/".format(number) + today + ".csv"
-        df = pd.DataFrame([])
-        self.lc_upload_to_s3(df, bucket_name, file_name) ##### change this to SageMaker
-        print("Finished the Lambda {} Process...".format(number))
-        print()
 
     # PYTRENDS FUNCTION THAT GETS TRENDING CITIES BASED ON KEYWORD(BRAND)
     def get_google_city(self, keyword):
@@ -248,9 +244,9 @@ class data_extraction:
         """gets our acv curves from s3 in order to run the distributor path option for upcs
         Returns:
              df: acv_df is the accounts with the respective net list running percent for each category and tier 
-        """
-        bucket='ejgallo-lake-sales-dev'
-        key = '2021/benchmark/data/acv_curves.csv'
+        """ 
+        bucket=bucket_name
+        key = benchmark_acv_str
         obj = s3c.get_object(Bucket= bucket , Key = key)
         acv_df = pd.read_csv(io.BytesIO(obj['Body'].read()), encoding='utf8',low_memory = False)
         columns = ['9L VOLUME','PHYS VOLUME','NET LIST DOLLARS']
@@ -265,8 +261,7 @@ class data_extraction:
         acv_df['concat'] = acv_df['concat'].astype(str)
         unique_citystates = acv_df[['Mkt_Grp_State','Acct_City','concat']].drop_duplicates()
         # Read in geocodes and fuzzy match them--> need to do ts cause some cities have different spellings from the hana side
-        bucket='ejgallo-lake-sales-dev'
-        key = '2021/benchmark/data/geo_codes.csv'
+        key = benchmark_geo_str
         obj = s3c.get_object(Bucket= bucket , Key = key)
         geo_codes = pd.read_csv(io.BytesIO(obj['Body'].read()), encoding='utf8',low_memory = False)
         geo_codes['concat'] =geo_codes['Mkt_Grp_State'] + ' ' + geo_codes['City']
@@ -293,8 +288,8 @@ class data_extraction:
         Returns:
             [df]: [upc_df is all upcs and their corresponding price tier and category: to be merged onto acv df]
         """
-        bucket='ejgallo-lake-sales-dev'
-        key = '2021/benchmark/data/UPC_Tier_Dict.csv'
+        bucket=bucket_name
+        key = upc_tier_str
         obj = s3c.get_object(Bucket= bucket , Key = key)
         upc_df = pd.read_csv(io.BytesIO(obj['Body'].read()), encoding='utf8',low_memory = False)
         # Add leading zeros to UPC to match Gallo data
@@ -333,7 +328,7 @@ class prime_directive:
     def __init__(self, change_names = {'FINE_WINE':'Fine_Wine_Acct_Flag', 'INFLUENCER':'Fine_Wine_Inflncr_Acct_Flag','ICON':'Fine_Wine_Icon_Acct_Flag'}):
         self.change_names = change_names
 
-    def search_func(self, upc_value, platform='cloud', bucket_name='ejgallo-lake-sales-dev', file_name='2021/prime_directive/off_brand_standards_Q2.csv'):
+    def search_func(self, upc_value, bucket_name, file_name, platform='cloud'):
         de = data_extraction()
         if platform == "sagemaker":
             brand_standards = de.sg_reading_in_from_s3(bucket_name, file_name,dtypes = {'UPC':str})
@@ -411,7 +406,7 @@ class prime_directive:
             store_list = ['GROCERY','LIQUOR','CONVENIENCE','DRUG','MASS MERCHANDISER','ALL OTHER-OFF SALE','DOLLAR','CLUB']
             for upc in upc_list:
                 try:
-                    search_dict = self.search_func(upc)
+                    search_dict = self.search_func(upc, bucket_name, brand_standards_str)
                     upc_df = df[df['UPC']==upc]
                     upc_df = upc_df.rename(columns = self.change_names)
                     upc_df.columns = map(str.lower, upc_df.columns)
@@ -425,7 +420,7 @@ class prime_directive:
                                                                   x['fine_wine_acct_flag'],x['gsp']), axis = 1)
                     upc_df = upc_df[upc_df['standards_check'] == 'Y']
                 except Exception as err:
-                    # print('passing this upc {}... {}...'.format(upc, err))
+                    print('passing this upc {}... {}...'.format(upc, err))
                     continue
                 brand_standard_list.append(upc_df)
             try:
@@ -471,8 +466,6 @@ class data_conversion:
     def __init__(self, categorical_list, numerical_list):
         self.categorical_list = categorical_list
         self.numerical_list = numerical_list
-        self.bucket_name = 'ejgallo-lake-sales-dev'
-        self.dictionary_name = 'prophet/model_weights/labels'
 
     def s3_connection(self):
         s3c = boto3.client(
@@ -481,13 +474,11 @@ class data_conversion:
             aws_access_key_id = cred.AWS_ACCESS_KEY,
             aws_secret_access_key = cred.AWS_SECRET_KEY
             )
-        return s3c        
+        return s3c
 
-    def load_labels(self, label_name):
-        bucket_name = self.bucket_name
-        directory_name = self.dictionary_name
+    def load_labels(self, bucket_name, directory_name, label_name):
         s3c = self.s3_connection()
-        KEY = '{}/labels_{}.pickle'.format(self.dictionary_name, label_name)
+        KEY = '{}/labels_{}.pickle'.format(directory_name, label_name)
         try:
             obj = s3c.get_object(Bucket=bucket_name, Key = KEY)
             label_dict = pickle.loads(obj['Body'].read())
@@ -497,11 +488,9 @@ class data_conversion:
             label_dict = {}
         return label_dict
 
-    def save_labels(self, label_dict, label_name):
-        bucket_name = self.bucket_name
-        directory_name = self.dictionary_name
+    def save_labels(self, label_dict, bucket_name, directory_name, label_name):
         s3c = self.s3_connection()
-        KEY = '{}/labels_{}.pickle'.format(self.dictionary_name, label_name)
+        KEY = '{}/labels_{}.pickle'.format(directory_name, label_name)
         try:
             serialized_df = pickle.dumps(label_dict)
             s3c.put_object(Bucket=bucket_name, Key=KEY, Body=serialized_df)
@@ -511,28 +500,6 @@ class data_conversion:
             print("Error while saving a label file for {}...".format(str(err)))
             print()         
 
-    def data_conversion_cat(self, df, label_name):
-        df_num = df[self.numerical_list]
-        label_dict = self.load_labels(label_name)
-        if self.categorical_list != []:
-            print("Detected categorical list for sustainable growth...")
-            df_cat = df[self.categorical_list]
-            if label_dict == {}:
-                for category in self.categorical_list:
-                    label_encoder = LabelEncoder()
-                    df_cat.loc[:,category] = label_encoder.fit_transform(df_cat.loc[:,category]) # use label_encoder.inverse_transform(data) to go back to original labels
-                    label_dict[category] = label_encoder
-                self.save_labels(label_dict, label_name)
-            else:
-                for category in self.categorical_list:
-                    label_encoder = label_dict[category]
-                    df_cat.loc[:,category] = label_encoder.transform(df_cat.loc[:,category])
-            df = pd.concat([df_cat, df_num], axis=1)
-        else:
-            print("There is no categorical list for sustainable growth...")
-            df = df_num
-        return df, label_dict
-
     def data_conversion_cat_onehot(self, df):
         df_num = df[self.numerical_list]
         if self.categorical_list != []:
@@ -543,19 +510,51 @@ class data_conversion:
             df = df_num
         return df
 
-    def data_conversion_num(self, df, scaler_type = "minmax"):
+    def data_conversion_cat(self, df, bucket_name, directory_name, label_name):
+        df_num = df[self.numerical_list]
+        label_dict = self.load_labels(bucket_name, directory_name, label_name)
+        if self.categorical_list != []:
+            print("Detected categorical list for sustainable growth...")
+            df_cat = df[self.categorical_list]
+            if label_dict == {}:
+                for category in self.categorical_list:
+                    label_encoder = LabelEncoder()
+                    df_cat.loc[:,category] = label_encoder.fit_transform(df_cat.loc[:,category]) # use label_encoder.inverse_transform(data) to go back to original labels
+                    label_dict[category] = label_encoder
+                self.save_labels(label_dict,  bucket_name, directory_name, label_name)
+            else:
+                for category in self.categorical_list:
+                    label_encoder = label_dict[category]
+                    df_cat.loc[:,category] = label_encoder.transform(df_cat.loc[:,category])
+            df = pd.concat([df_cat, df_num], axis=1)
+        else:
+            print("There is no categorical list for the data...")
+            df = df_num
+        return df, label_dict
+
+    def data_conversion_num(self, df, bucket_name, directory_name, label_name, scaler_type = "minmax"):
         categorical_list = list(df.columns)
         categorical_list = [category for category in categorical_list if category not in self.numerical_list]
         df_cat = df[categorical_list]
         df_num = df[self.numerical_list]
-        for variable in self.numerical_list:
-            if scaler_type == "minmax":
-                scaler = MinMaxScaler()
-            elif scaler_type == "power":
-                scaler = PowerTransformer()
+        label_dict = self.load_labels(bucket_name, directory_name, label_name)
+        if self.numerical_list != []:
+            print("Detected numerical list for the data...")
+            if label_dict == {}:
+                if scaler_type == "minmax":
+                    scaler = MinMaxScaler()
+                elif scaler_type == "power":
+                    scaler = PowerTransformer()
+                else:
+                    scaler = StandardScaler()               
+                scaler.fit_transform(df_num)
+                self.save_labels(scaler,  bucket_name, directory_name, label_name)
             else:
-                scaler = StandardScaler()
-            df_num.loc[:,variable] = scaler.fit_transform(np.array(df_num.loc[:,variable]).reshape(-1, 1))
+                scaler = label_dict
+                scaler.transform(df_num)
+        else:
+            print("There is no numerical list for sustainable growth...")
+            df = df_cat
         df = pd.concat([df_cat, df_num], axis=1)
         return df
 
@@ -638,9 +637,9 @@ class data_visualization:
                 i = 0
                 j += 1
 
-    def elbow_graph(self, df, model_name='KMEANS', range_start=1, range_end=11):
+    def elbow_graph(self, df, model_name='KMEANS', range_start=2, range_end=11):
         distortions = []
-        for i in range(1,11):
+        for i in range(range_start, range_end):
             print("Fitting cluster {}...".format(i))
             if model_name == "KMEANS":
                 km = KMeans(n_clusters=i, init='random',n_init=10,max_iter=100,tol=1e-04)
@@ -687,11 +686,11 @@ class data_visualization:
 
     def umap_graph(self, df):
         numerical_df = df.select_dtypes(exclude='object')
-        for c in numerical.columns:
+        for c in numerical_df.columns:
             pt = PowerTransformer()
             numerical_df.loc[:, c] = pt.fit_transform(np.array(numerical_df[c]).reshape(-1, 1))
         categorical_df = df.select_dtypes(include='object')
-        categorical_df = pd.get_dummies(categorical)
+        categorical_df = pd.get_dummies(categorical_df)
         categorical_weight = len(df.select_dtypes(include='object').columns) / len(df.columns)
         fit1 = umap.UMAP(metric='l2').fit(numerical_df)
         fit2 = umap.UMAP(metric='dice').fit(categorical_df)
@@ -710,12 +709,9 @@ class data_visualization:
 ########################### Model Training ###########################
 class model_training:
     def __init__(self):
-        self.bucket_name = 'ejgallo-lake-sales-dev'
-        self.dictionary_name = 'prophet/model_weights/models'
+        pass
 
-    def load_model_trained(self, model_name, label_name):
-        bucket_name = self.bucket_name
-        directory_name = self.dictionary_name
+    def load_model_trained(self, bucket_name, directory_name, model_name, label_name):
         s3c = boto3.client(
                 's3', 
                 region_name = cred.AWS_REGION_NAME,
@@ -730,9 +726,7 @@ class model_training:
         except Exception as err:
             print("Loading failed for {}...".format(str(err)))
             print()
-    def save_model_trained(self, model_trained, model_name, label_name):
-        bucket_name = self.bucket_name
-        directory_name = self.dictionary_name
+    def save_model_trained(self, model_trained, bucket_name, directory_name, model_name, label_name):
         s3c = boto3.client(
                 's3', 
                 region_name = cred.AWS_REGION_NAME,
@@ -877,6 +871,92 @@ class model_training:
         model_trained.fit(df, categorical=categorical_list)
         return model_trained
 
+########################### Polygon Map ###########################
+class polygon_map:
+    def __init__(self):
+        pass
+    def Left_index(self, points):
+        minn = 0
+        for i in range(1,len(points)):
+            if points[i].x < points[minn].x:
+                minn = i
+            elif points[i].x == points[minn].x:
+                if points[i].y > points[minn].y:
+                    minn = i
+        return minn
+
+    def orientation(self, p, q, r):
+        val = (q.y - p.y) * (r.x - q.x) - \
+            (q.x - p.x) * (r.y - q.y)
+        if val == 0:
+            return 0
+        elif val > 0:
+            return 1
+        else:
+            return 2
+
+    def convexHull(self, points, n):
+        if n < 3:
+            return
+        l = self.Left_index(points)
+        hull = []
+        p = l
+        q = 0
+        while(True):
+            hull.append(p)
+            q = (p + 1) % n
+            for i in range(n):
+                if(self.orientation(points[p],
+                            points[i], points[q]) == 2):
+                    q = i
+            p = q
+            if(p == l):
+                break
+        lon = []
+        lat = []
+        for each in hull:
+            lon.append(points[each].x)
+            lat.append(points[each].y)
+        return lon, lat
+
+    def cluster_creator(self, df, index_list, eps, min_samples, metric, metric_params, algorithm, leaf_size, p, n_jobs):
+        model = DBSCAN(eps=eps, min_samples=min_samples, metric=metric, metric_params=metric_params, algorithm=algorithm, leaf_size=leaf_size, p=p, n_jobs=n_jobs)
+        prediction = model.fit_predict(df[index_list])
+        pred = pd.DataFrame(prediction, columns=['cluster_group'])
+        df = df.join(pred)
+        return df
+
+    def polygon_creator(self, df, longitude, latitude):
+        lon = df[longitude]
+        lat = df[latitude]
+        df = GeoDataFrame(df, geometry=points_from_xy(lon, lat))
+        df.rename(columns={"geometry": "points"}, inplace=True)
+        points = [Point(x,y) for x,y in zip(lon, lat)]
+        new_lon, new_lat = self.convexHull(points, len(points))
+        polygon_geom = [Polygon(zip(new_lon, new_lat))]*len(df)
+        crs = 'epsg:4326'
+        df = GeoDataFrame(df, crs=crs, geometry=polygon_geom)  
+        return df
+        
+    def map_creator(self, df, index_list, eps=0.5, min_samples=5, metric='euclidean', metric_params=None, algorithm='auto', leaf_size=30, p=None, n_jobs=None):
+        df = self.cluster_creator(df, index_list, eps=eps, min_samples=min_samples, metric=metric, metric_params=metric_params, algorithm=algorithm, leaf_size=leaf_size, p=p, n_jobs=n_jobs)
+        n_clusters = len(set(df['cluster_group']))
+        df_list = []
+        for cluster in range(n_clusters):
+            part_df = df[df['cluster_group']==cluster]
+            part_df = self.polygon_creator(part_df, 'Long', 'Lat')
+            df_list.append(part_df)
+        total_df = pd.concat(df_list, ignore_index=False)
+        return total_df
+
+    def boundary_test(self, acct_name, polygon, points):
+        if polygon.contains(points):
+            print("{} is in the polygon...".format(acct_name))
+        elif polygon.touches(points):
+            print("{} is in the polygon...".format(acct_name))
+        else:
+            print("Not in the polygon...")
+
 
 ########################### Time Series Models ###########################
 class time_series_models:
